from typing import Any, Tuple

import torch
from torch.autograd import Function

from pipegoose.distributed.functional import all_gather, all_reduce, scatter
from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.distributed.parallel_mode import ParallelMode


class _Broadcast(Function):
    @staticmethod
    def forward(ctx, tensor: torch.Tensor, parallel_context: ParallelContext) -> torch.Tensor:
        ctx.parallel_context = parallel_context

        return tensor

    @staticmethod
    def backward(ctx: Any, grad: torch.Tensor) -> Tuple[torch.Tensor, None, None]:
        parallel_context = ctx.parallel_context

        all_reduce(grad, parallel_context=parallel_context, parallel_mode=ParallelMode.TENSOR)

        return (grad, None, None)


class _Gather(Function):
    @staticmethod
    def forward(ctx: Any, input: torch.Tensor, dim: int, parallel_context: ParallelContext) -> torch.Tensor:
        ctx.dim = dim
        ctx.parallel_context = parallel_context

        return all_gather(input, dim=dim, async_op=False, parallel_context=parallel_context, parallel_mode=ParallelMode.TENSOR)

    @staticmethod
    def backward(ctx: Any, grad: torch.Tensor) -> Tuple[torch.Tensor, None, None]:
        dim = ctx.dim
        parallel_context = ctx.parallel_context

        return (
            scatter(grad, dim=dim, parallel_context=parallel_context, parallel_mode=ParallelMode.TENSOR),
            None,
            None,
        )


class _Scatter(Function):
    @staticmethod
    def forward(ctx: Any, input: torch.Tensor, dim: int, parallel_context: ParallelContext) -> torch.Tensor:
        ctx.dim = dim
        ctx.parallel_context = parallel_context
        return scatter(input, dim=dim, parallel_context=parallel_context, parallel_mode=ParallelMode.TENSOR)

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> Tuple[torch.Tensor, None, None]:
        dim = ctx.dim
        parallel_context = ctx.parallel_context

        return (
            all_gather(
                grad_output, dim=dim, async_op=False, parallel_context=parallel_context, parallel_mode=ParallelMode.TENSOR
            ),
            None,
            None,
        )


class _Reduce(Function):
    @staticmethod
    def forward(ctx: Any, input: torch.Tensor, parallel_context: ParallelContext) -> torch.Tensor:
        return all_reduce(input, parallel_context=parallel_context, parallel_mode=ParallelMode.TENSOR)

    @staticmethod
    def backward(ctx: Any, grad: torch.Tensor) -> Tuple[torch.Tensor, None]:
        return (grad, None)


def broadcast_to_tensor_group(input: torch.Tensor, parallel_context: ParallelContext):
    return _Broadcast.apply(input, parallel_context)


def gather_to_tensor_group(input: torch.Tensor, dim: int, parallel_context: ParallelContext):
    return _Gather.apply(input, dim, parallel_context)


def scatter_to_tensor_group(input, dim, parallel_context):
    return _Scatter.apply(input, dim, parallel_context)


def reduce_to_tensor_group(input, parallel_context):
    return _Reduce.apply(input, parallel_context)



from queue import Queue
from typing import List, Optional, Tuple

import torch
from torch import nn

from pipegoose.nn.pipeline_parallel._dependency import create_backward_dependency
from pipegoose.nn.pipeline_parallel._worker import Task, spawn_worker
from pipegoose.nn.pipeline_parallel.microbatch import Batch
from pipegoose.nn.pipeline_parallel.scheduler import BaseScheduler, DetermisticScheduler


class Pipeline:
    """A base class for pipeline."""

    def __init__(
        self,
        batches: List[Batch],
        partitions: List[nn.Sequential],
        devices: Optional[List[torch.device]] = None,
        scheduler: BaseScheduler = DetermisticScheduler(),
    ) -> None:
        """Initialize the pipeline.

        Args:
            batches (List[Batch]): A list of micro-batches.
            partitions (List[nn.Sequential]): A partitioned model.
            devices (Optional[List[torch.device]], optional): A list of devices. Defaults to None.
            scheduler (BaseScheduler, optional): _description_. Defaults to DetermisticScheduler().
        """
        self.batches = batches
        self.partitions = partitions
        self.devices = devices
        self.scheduler = scheduler

    def fit(self):
        batches = self.batches
        partitions = self.partitions
        devices = self.devices
        scheduler = self.scheduler

        n_batches = len(batches)
        n_partitions = len(partitions)

        with spawn_worker(devices) as (in_queues, out_queues):
            for schedule in scheduler.generate(n_batches, n_partitions):
                self._depend(schedule)
                self._compute(schedule, in_queues, out_queues)

    def _depend(self, schedule: List[Tuple[int, int]]):
        """Enforce the dependency between batches and partitions."""
        batches = self.batches

        for microbatch_idx, partition_idx in schedule:
            if microbatch_idx != 0:
                create_backward_dependency(batches[microbatch_idx - 1], batches[microbatch_idx])

    def _compute(self, schedule: List[Tuple[int, int]], in_queues: List[Queue], out_queues: List[Queue]):
        """Compute the partitions."""
        batches = self.batches
        partitions = self.partitions

        for microbatch_idx, partition_idx in schedule:
            batch = batches[microbatch_idx]
            partrition = partitions[partition_idx]

            def compute(batch, partrition):
                def wrapper():
                    return partrition(batch)

                return wrapper

            task = Task(compute=compute(batch, partrition))
            in_queues[partition_idx].put(task)

        for microbatch_idx, partition_idx in schedule:
            queue_output = out_queues[partition_idx].get()
            task, output = queue_output.task, queue_output.output

            # put the output back to the batch
            batches[microbatch_idx] = output



from abc import ABC, abstractclassmethod
from typing import Annotated, Iterable, List, Tuple


class BaseScheduler(ABC):
    @abstractclassmethod
    def generate(self):
        raise NotImplementedError


class DetermisticScheduler(BaseScheduler):
    """
    torchgpipe: On-the-fly Pipeline Parallelism for Training Giant Models
    https://arxiv.org/abs/2004.09910

    Section 3.2.1: Forward Dependency: Deterministic Clock-cycle
    """

    def generate(
        self,
        n_microbatches: int,
        n_partitions: int,
    ) -> Iterable[List[Tuple[Annotated[int, "batch_idx"], Annotated[int, "partition_idx"]]]]:
        assert (
            n_microbatches > 0
        ), "The number of microbatches must be \
            greater than 0"
        assert (
            n_partitions > 0
        ), "The number of partitions must be \
            greater than 0"

        self.n_partitions = n_partitions
        self.n_microbatches = n_microbatches
        n_clock_cycles = self.n_partitions + self.n_microbatches - 1

        for clock_idx in range(n_clock_cycles):
            start_partrition = max(clock_idx + 1 - self.n_microbatches, 0)
            end_partition = min(clock_idx + 1, self.n_partitions)

            tasks = []
            for partition_idx in range(start_partrition, end_partition):
                microbatch_idx = clock_idx - partition_idx
                tasks.append((microbatch_idx, partition_idx))

            yield tasks



import random
import socket
from functools import partial
from typing import Callable

import torch.multiprocessing as mp


def find_free_port(min_port: int = 2000, max_port: int = 65000) -> int:
    while True:
        port = random.randint(min_port, max_port)
        try:
            with socket.socket() as sock:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.bind(("localhost", port))
                return port
        except OSError:
            continue


def spawn(func: Callable, world_size: int = 1, **kwargs):
    if kwargs.get("port") is None:
        port = find_free_port()
    else:
        port = kwargs["port"]
        kwargs.pop("port")

    wrapped_func = partial(func, world_size=world_size, port=port, **kwargs)
    mp.spawn(wrapped_func, nprocs=world_size)



from contextlib import contextmanager
from dataclasses import dataclass
from queue import Queue
from threading import Thread
from typing import Annotated, Any, Dict, Generator, List, Tuple

import torch

from pipegoose.nn.pipeline_parallel._task import Task


class InQueueTask:
    pass


class OutQueueTask:
    pass


@dataclass
class QueueOutput:
    task: Task
    output: Any
    is_done: bool = False


def wait_and_execute(device: torch.device, in_queue: Queue, out_queue: Queue):
    """Wait for a task and execute it."""
    while True:
        task = in_queue.get()

        if task.is_done is True:
            break

        try:
            output = task.compute()
        except Exception:
            out_queue.put(QueueOutput(task=task, output=None, is_done=False))
            continue

        out_queue.put(QueueOutput(task=task, output=output, is_done=True))


@contextmanager
def spawn_worker(
    devices: List[torch.device],
) -> Generator[
    Tuple[
        Annotated[List[Queue], "A list of tasks to be executed"],
        Annotated[List[Queue], "A list of tasks has been executed"],
    ],
    None,
    None,
]:
    """Spawn new worker threads."""
    in_queues: List[Queue] = []
    out_queues: List[Queue] = []

    workers: Dict[torch.device, Tuple[Queue, Queue]] = {}

    for device in devices:
        try:
            in_queue, out_queue = workers[device]
        except KeyError:
            in_queue = Queue()
            out_queue = Queue()
            workers[device] = (in_queue, out_queue)

            thread = Thread(target=wait_and_execute, args=(device, in_queue, out_queue), daemon=True)
            thread.start()

        in_queues.append(in_queue)
        out_queues.append(out_queue)

    yield (in_queues, out_queues)


# class Worker:
#     def __init__(self, devices: List[torch.device]):
#         self.devices = devices
#         self._streams = [torch.cuda.Stream(device=device) for device in devices]
#         self.in_queues: List[Queue] = []
#         self.out_queues: List[Queue] = []

#     @contextmanager
#     def spawn(self) -> Generator[Tuple[Queue, Queue]]:
#         pass



from typing import Any

import torch
import torch.distributed as dist

from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.distributed.parallel_mode import ParallelMode

ID_TO_DTYPE = [
    torch.bfloat16,
    torch.float16,
    torch.float32,
    torch.float64,
    torch.uint8,
    torch.int8,
    torch.int16,
    torch.int32,
    torch.int64,
    torch.bool,
]

DTYPE_TO_ID = {dtype: idx for idx, dtype in enumerate(ID_TO_DTYPE)}


class _P2P:
    def __init__(self):
        self._INSTRUCTIONS = {
            torch.Tensor: {"send": self._send_tensor, "recv": self._recv_tensor},
        }

    def _send_metadata(
        self,
        data: torch.Tensor,
        dst: int,
        parallel_context: ParallelContext,
        parallel_mode: ParallelMode,
    ):
        assert isinstance(data, torch.Tensor), "data must be a torch.Tensor"

        group = parallel_context.get_group(parallel_mode)

        dtype = torch.tensor(DTYPE_TO_ID[data.dtype])
        dist.send(dtype, dst=dst, group=group)

        requires_grad = torch.tensor(1 if data.requires_grad else 0)
        dist.send(requires_grad, dst=dst, group=group)

        shape = torch.tensor(list(data.shape))
        dist.send(shape, dst=dst, group=group)

    def _recv_metadata(
        self,
        src: int,
        parallel_context: ParallelContext,
        parallel_mode: ParallelMode,
    ):
        group = parallel_context.get_group(parallel_mode)

        dtype = torch.tensor(0)
        dist.recv(dtype, src=src, group=group)
        dtype = ID_TO_DTYPE[dtype.item()]

        requires_grad = torch.tensor(0)
        dist.recv(requires_grad, src=src, group=group)
        requires_grad = True if requires_grad == 1 else False

        shape = torch.tensor(0)
        dist.recv(shape, src=src, group=group)
        if isinstance(shape.tolist(), int):
            shape = (shape.item(),)
        else:
            shape = tuple(shape.tolist())

        return dtype, requires_grad, shape

    def _send_tensor(
        self,
        data: torch.Tensor,
        dst: int,
        parallel_context: ParallelContext,
        parallel_mode: ParallelMode,
    ):
        assert isinstance(data, torch.Tensor), "data must be a torch.Tensor"

        self._send_metadata(data, dst, parallel_context, parallel_mode)

        group = parallel_context.get_group(parallel_mode)
        dist.send(data, dst=dst, group=group)

    def _recv_tensor(self, src: int, parallel_context: ParallelContext, parallel_mode: ParallelMode) -> torch.Tensor:
        group = parallel_context.get_group(parallel_mode)

        dtype, requires_grad, shape = self._recv_metadata(src, parallel_context, parallel_mode)

        data = torch.zeros(size=shape, dtype=dtype, requires_grad=requires_grad)
        dist.recv(data, src=src, group=group)

        return data

    def send(self, data: Any, dst: int, parallel_context: ParallelContext, parallel_mode: ParallelMode):
        _type = type(data)
        assert _type in self._INSTRUCTIONS, f"Type {_type} is not supported"

        self._INSTRUCTIONS[_type]["send"](data, dst, parallel_context, parallel_mode)

    def recv(self, src: int, parallel_context: ParallelContext, parallel_mode: ParallelMode) -> torch.Tensor:
        # TODO: Add support for other types
        return self._INSTRUCTIONS[torch.Tensor]["recv"](src, parallel_context, parallel_mode)



import time
from abc import ABC, abstractclassmethod
from typing import List

import torch
from torch import nn

from pipegoose.nn.pipeline_parallel.batch import Batch


class BasePartitioner(ABC):
    @abstractclassmethod
    def split(self):
        raise NotImplementedError


class TimePartritioner(BasePartitioner):
    def __init__(self, model: nn.Module, devices: List[torch.device]):
        self.model = model

    def split(self, data):
        for layer in self.model.named_children():
            pass


def profile_elasped_time_per_layer(model: nn.Module, batches: List[Batch]) -> List[float]:
    records = []

    for layer in model.children():
        # start_event = torch.cuda.Event(enable_timing=True)
        # end_event = torch.cuda.Event(enable_timing=True)
        start_time = time.time()

        # start_event.record()
        outputs = [layer(batch) for batch in batches]
        output_with_grad = [x for x in outputs if x.requires_grad]

        if len(output_with_grad) > 0:
            for output in output_with_grad:
                torch.autograd.backward(output, output)

        # end_event.record()
        end_time = time.time()
        records.append(end_time - start_time)



from typing import Any, Dict

import torch
from torch import nn

from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.nn.data_parallel.sharding import GreedySharding, ShardingStategy


class FullyShardedDataParallel(nn.Module):
    def __init__(
        self,
        module: nn.Module,
        sharding: ShardingStategy = GreedySharding(),
        parallel_context: ParallelContext = None,
    ):
        super().__init__()

        self.module = sharding.shard(module, parallel_context)

    def forward(self, *args: Any, **kwargs: Any) -> Any:
        self._rebuild_full_params()
        output = self.module(**args, **kwargs)

        return output

    def _lazy_init(self):
        self._init_streams()

    def _setup_streams(self):
        if torch.cuda.is_available():
            self._streams: Dict[str, torch.cuda.Stream] = {}
            self._streams["all_gather"] = torch.cuda.Stream()
            self._streams["all_reduce"] = torch.cuda.Stream()

    def pre_forward_hook(self):
        pass

    def post_forward_hook(self):
        pass

    def _rebuild_full_params(self):
        pass

    def _free_full_params(self):
        pass



from typing import Union

import torch


class CPUStreamType:
    pass


StreamType = Union[CPUStreamType, torch.cuda.Stream]


def is_cuda(stream: StreamType) -> bool:
    return True if isinstance(stream, torch.cuda.Stream) else False


def wait_stream(source: StreamType, target: StreamType):
    """Source stream waits for target stream completes it queues."""
    if isinstance(target, torch.cuda.Stream):
        if isinstance(source, torch.cuda.Stream):
            source.wait_stream(target)
        else:
            torch.cuda.synchronize()


def default_stream(device: torch.device) -> StreamType:
    """Get the default stream of the device."""
    if device.type == "cpu":
        return CPUStreamType()
    elif device.type == "cuda":
        return torch.cuda.default_stream(device)
    else:
        raise ValueError(f"Unknown device type: {device.type}")


def get_device(stream: StreamType) -> torch.device:
    """Get the device of the stream."""
    if isinstance(stream, torch.cuda.Stream):
        return stream.device
    return torch.device("cpu")


def use_stream(stream: StreamType):
    if not is_cuda(stream):
        yield
        return

    with torch.cuda.stream(stream):
        yield


def record_stream(tensor: torch.Tensor, stream: StreamType) -> None:
    if is_cuda(stream):
        tensor.record_stream(stream)



import torch.distributed as dist

from pipegoose.distributed._initializers.initializer import (
    ProcessGroupInitializer,
    ProcessGroupResult,
)
from pipegoose.distributed.parallel_mode import ParallelMode


class PipelineParallelGroupInitializer(ProcessGroupInitializer):
    def init_dist_group(self) -> ProcessGroupResult:
        num_pipeline_parallel_groups = self.world_size // self.pipeline_parallel_size
        local_rank = None
        local_world_size = None
        ranks_in_group = None
        process_group = None
        parallel_mode = ParallelMode.PIPELINE

        for i in range(num_pipeline_parallel_groups):
            ranks = list(range(i, self.world_size, num_pipeline_parallel_groups))

            # NOTE: dist.new_group() must be called collectively by all the processes
            # that would be part of the group, which means every process in the group
            # needs to call this function. If only a subset of the processes call new_group(),
            # it will hang because it's waiting for the rest of the processes to join.
            process_group = dist.new_group(ranks=ranks)

            if self.rank in ranks:
                local_rank = ranks.index(self.rank)
                local_world_size = len(ranks)
                ranks_in_group = ranks

        return {
            "local_rank": local_rank,
            "process_group": process_group,
            "local_world_size": local_world_size,
            "ranks_in_group": ranks_in_group,
            "parallel_mode": parallel_mode,
        }



from typing import Any, Optional

import torch
import torch.distributed as dist
from torch.distributed import ReduceOp

from pipegoose.distributed._p2p import _P2P
from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.distributed.parallel_mode import ParallelMode


def scatter(
    tensor: torch.Tensor,
    dim: int,
    parallel_context: Optional[ParallelContext] = None,
    parallel_mode: Optional[ParallelMode] = None,
) -> torch.Tensor:
    world_size = parallel_context.get_world_size(parallel_mode)
    rank = parallel_context.get_local_rank(parallel_mode)

    if world_size == 1:
        return tensor

    assert tensor.size(dim) % world_size == 0

    tensor_list = torch.chunk(tensor, world_size, dim=dim)
    return tensor_list[rank]


def reduce(
    tensor: torch.Tensor,
    dst: int,
    op: ReduceOp = ReduceOp.SUM,
    async_op: bool = False,
    parallel_context: Optional[ParallelContext] = None,
    parallel_mode: Optional[ParallelContext] = None,
) -> torch.Tensor:
    world_size = parallel_context.get_world_size(parallel_mode)

    if world_size == 1:
        return tensor

    group = parallel_context.get_group(parallel_mode)
    work = dist.reduce(tensor, dst=dst, op=op, group=group)

    if async_op:
        return tensor, work
    else:
        return tensor


def broadcast(
    tensor: torch.Tensor,
    src: int,
    async_op: bool = False,
    parallel_context: Optional[ParallelContext] = None,
    parallel_mode: Optional[ParallelMode] = None,
) -> torch.Tensor:
    world_size = parallel_context.get_world_size(parallel_mode)

    if world_size == 1:
        return tensor

    group = parallel_context.get_group(parallel_mode)
    work = dist.broadcast(tensor, src=src, group=group, async_op=async_op)

    if async_op:
        return tensor, work
    else:
        return tensor


def all_gather(
    tensor: torch.Tensor,
    dim: int = 0,
    async_op: bool = False,
    parallel_context: Optional[ParallelContext] = None,
    parallel_mode: Optional[ParallelMode] = None,
) -> torch.Tensor:
    """All gather tensors from all processes in parallel group.

    Args:
        input (torch.Tensor): The tensor you want to gather.
        dim (int, optional): The dimension along which to gather the tensor.. Defaults to 0.
        async_op (bool, optional): _description_. Defaults to False.
        parallel_context (Optional[ParallelContext], optional): _description_. Defaults to None.
        parallel_mode (Optional[ParallelMode], optional): _description_. Defaults to None.

    Returns:
        torch.Tensor: _description_
    """
    world_size = parallel_context.get_world_size(parallel_mode)
    group = parallel_context.get_group(parallel_mode)

    if world_size == 1:
        return tensor

    tensor_list = [torch.empty_like(tensor) for _ in range(world_size)]
    work = dist.all_gather(tensor_list=tensor_list, tensor=tensor, async_op=async_op, group=group)

    if tensor.dim() == 0:
        tensor_list = [tensor.unsqueeze(dim=0) for tensor in tensor_list]

    tensor_list = torch.cat(tensor_list, dim=dim)

    if async_op:
        return tensor_list, work
    else:
        return tensor_list


def all_reduce(
    tensor: torch.Tensor,
    op: ReduceOp = ReduceOp.SUM,
    async_op: bool = False,
    parallel_context: Optional[ParallelContext] = None,
    parallel_mode: Optional[ParallelMode] = None,
) -> torch.Tensor:
    world_size = parallel_context.get_world_size(parallel_mode)

    if world_size == 1:
        return tensor

    group = parallel_context.get_group(parallel_mode)
    work = dist.all_reduce(tensor, op=op, group=group, async_op=async_op)

    if async_op:
        return tensor, work
    else:
        return tensor


def reduce_scatter():
    pass


def send(
    data: Any,
    src: int,
    dst: int,
    parallel_context: ParallelContext,
    parallel_mode: ParallelMode = ParallelMode.PIPELINE,
):
    if src == parallel_context.get_local_rank(parallel_mode):
        _P2P().send(data, dst, parallel_context, parallel_mode)


def recv(
    src: int, dst: int, parallel_context: ParallelContext, parallel_mode: ParallelMode = ParallelMode.PIPELINE
) -> Optional[Any]:
    if dst == parallel_context.get_local_rank(parallel_mode):
        return _P2P().recv(src, parallel_context, parallel_mode)



from dataclasses import dataclass

import torch


@dataclass
class Data:
    # data
    microbatch_idx: int
    partition_idx: int
    data: torch.Tensor  # the actual data

    # job
    is_forward: bool
    is_training: bool
    is_grad_enabled: bool  # not implemented yet
    is_fp16: bool  # not implemented yet

    # communication
    src: int
    dst: int



import torch


def _is_using_full_storage(input: torch.Tensor) -> bool:
    """Check if the tensor is using full storage.

    Args:
        tensor (torch.Tensor): a tensor

    Returns:
        bool: True if the tensor is using full storage, otherwise False
    """
    total_storage = input.storage().size()
    n_elements = input.numel()

    if total_storage == n_elements:
        return True
    else:
        return False


def free_storage(input: torch.Tensor) -> None:
    """Directly free the storage of a tensor.

    Args:
        input (torch.Tensor): a tensor
    """
    if input.storage().size() > 0:
        assert _is_using_full_storage(input) is True
        input.storage().resize_(0)



import os
import random
from typing import List, Literal

import torch
import torch.distributed as dist
import torch.distributed.rpc as rpc

from pipegoose.distributed._initializers.initialize_data import (
    DataParallelGroupInitializer,
)
from pipegoose.distributed._initializers.initialize_pipeline import (
    PipelineParallelGroupInitializer,
)
from pipegoose.distributed._initializers.initialize_tensor import (
    TensorParallelGroupInitializer,
)
from pipegoose.distributed.parallel_mode import ParallelMode

DistributedBackend = Literal["gloo", "mpi", "nccl"]


WORKER_NAME = "WORKER_{}"


class ParallelContext:
    """Inspired from OSLO's parallel context:
    https://github.com/EleutherAI/oslo/blob/f16c73bc5893cd6cefe65e70acf6d88428a324e1/oslo/torch/distributed/parallel_context.py#L53
    """

    @classmethod
    def from_torch(
        cls,
        seed: int,
        backend: DistributedBackend,
        tensor_parallel_size: int,
        pipeline_parallel_size: int,
        data_parallel_size: int,
    ):
        rank = int(os.environ["RANK"])
        local_rank = int(os.environ["LOCAL_RANK"])
        world_size = int(os.environ["WORLD_SIZE"])
        local_world_size = int(os.environ["LOCAL_WORLD_SIZE"])
        host = os.environ["MASTER_ADDR"]
        port = int(os.environ["MASTER_PORT"])

        return cls(
            rank=rank,
            local_rank=local_rank,
            world_size=world_size,
            local_world_size=local_world_size,
            host=host,
            port=port,
            seed=seed,
            backend=backend,
            tensor_parallel_size=tensor_parallel_size,
            pipeline_parallel_size=pipeline_parallel_size,
            data_parallel_size=data_parallel_size,
        )

    def __init__(
        self,
        rank: int,
        local_rank: int,
        world_size: int,
        local_world_size: int,
        host: str,
        port: int,
        seed: int,
        backend: DistributedBackend,
        tensor_parallel_size: int,
        pipeline_parallel_size: int,
        data_parallel_size: int,
    ):
        num_gpus_per_model = tensor_parallel_size * pipeline_parallel_size

        assert (
            world_size % data_parallel_size == 0
        ), "The total number of processes must be divisible by the data parallel size."
        assert world_size % num_gpus_per_model == 0, (
            "The total number of processes must be divisible by"
            "the number of GPUs per model (tensor_parallel_size * pipeline_parallel_size)."
        )
        assert num_gpus_per_model * data_parallel_size == world_size, (
            "The number of process requires to train all replicas",
            "must be equal to the world size.",
        )

        self.tensor_parallel_size = tensor_parallel_size
        self.pipeline_parallel_size = pipeline_parallel_size
        self.data_parallel_size = data_parallel_size

        self._global_ranks = {}
        self._local_ranks = {}
        self._world_sizes = {}
        self._groups = {}
        self._ranks_in_group = {}
        self._rank_to_devices = {}

        self.local_rank = local_rank
        self.local_world_size = local_world_size

        self.init_global_dist(rank, world_size, backend, host, port)
        self.init_parallel_groups()

        if torch.cuda.is_available():
            self.set_device()

        self.rpc_worker_map = {rank: WORKER_NAME.format(rank) for rank in self.get_ranks_in_group(ParallelMode.GLOBAL)}
        # TODO: add initialize from torch launcher
        self.init_rpc_workers(host, port)

        # self.set_seed(seed)

    def init_global_dist(self, rank: int, world_size: int, backend: DistributedBackend, host: str, port: int):
        """Initialize the global distributed group.

        Args:
            rank (int): global rank
            world_size (int): global world size
            backend (DistributedBackend): distributed backend
            host (str): communication host
            port (int): communication port
        """
        init_method = f"tcp://{host}:{port}"
        dist.init_process_group(
            rank=rank,
            world_size=world_size,
            backend=backend,
            init_method=init_method,
        )
        ranks = list(range(world_size))
        process_group = dist.new_group(
            ranks=ranks,
            backend=dist.get_backend(),
        )
        self._register_dist(rank, world_size, process_group, ranks_in_group=ranks, parallel_mode=ParallelMode.GLOBAL)
        self.add_global_rank(ParallelMode.GLOBAL, rank)

    def init_parallel_groups(self):
        rank = self.get_global_rank()
        world_size = self.get_world_size(ParallelMode.GLOBAL)

        # NOTE: ensure all processes have joined the global group
        # before creating other groups
        torch.distributed.barrier()

        params = {
            "rank": rank,
            "world_size": world_size,
            "tensor_parallel_size": self.tensor_parallel_size,
            "pipeline_parallel_size": self.pipeline_parallel_size,
            "data_parallel_size": self.data_parallel_size,
        }

        results = [
            TensorParallelGroupInitializer(**params).init_dist_group(),
            PipelineParallelGroupInitializer(**params).init_dist_group(),
            DataParallelGroupInitializer(**params).init_dist_group(),
        ]

        for result in results:
            self._register_dist(**result)

    def init_rpc_workers(self, host: str, port: int):
        if self.pipeline_parallel_size > 1:
            init_method = f"tcp://{host}:{port}"
            options = rpc.TensorPipeRpcBackendOptions(
                init_method=init_method,
            )

            rank = self.get_global_rank()
            ranks = self.get_ranks_in_group(ParallelMode.GLOBAL)
            world_size = self.get_world_size(ParallelMode.GLOBAL)

            if torch.cuda.is_available():
                for _rank in ranks:
                    if _rank == rank:
                        continue
                    options.set_device_map(WORKER_NAME.format(_rank), {rank: _rank})

            rpc.init_rpc(name=self.rpc_worker_map[rank], rank=rank, world_size=world_size, rpc_backend_options=options)

    def _register_dist(
        self,
        local_rank: int,
        local_world_size: int,
        process_group: dist.ProcessGroup,
        ranks_in_group: List[int],
        parallel_mode: ParallelMode,
    ):
        """Register distributed group based on the parallel mode.

        Args:
            local_rank (int): local rank
            local_world_size (int): local world size
            mode (ParallelMode): parallel mode
        """
        self.add_local_rank(parallel_mode, local_rank)
        self.add_world_size(parallel_mode, local_world_size)
        self.add_group(parallel_mode, process_group)
        # TODO: remove this
        self.add_ranks_in_group(parallel_mode, ranks_in_group)

    def set_device(self):
        num_devices_per_node = torch.cuda.device_count()
        if num_devices_per_node > 0:
            device = self.get_global_rank() % num_devices_per_node
            torch.cuda.set_device(device)

    def set_seed(self, seed: int):
        random.seed(seed)
        torch.manual_seed(seed)

        # TODO: set GPU seed
        if torch.cuda.is_available():
            # parallel_seed = seed
            pass

    def is_initialized(self, parallel_mode: ParallelMode) -> bool:
        """Check if the parallel mode is initialized.

        Args:
            mode (ParallelMode): parallel mode

        Returns:
            bool: True if the parallel mode is initialized, False otherwise
        """
        return True if parallel_mode in self._groups else False

    def get_global_rank(self) -> int:
        return self._global_ranks[ParallelMode.GLOBAL]

    def add_global_rank(self, parallel_mode: ParallelMode, rank: int):
        self._global_ranks[parallel_mode] = rank

    def get_local_rank(self, parallel_mode: ParallelMode) -> int:
        return self._local_ranks[parallel_mode]

    def add_local_rank(self, parallel_mode: ParallelMode, rank: int):
        self._local_ranks[parallel_mode] = rank

    def get_world_size(self, parallel_mode: ParallelMode) -> int:
        return self._world_sizes[parallel_mode]

    def add_world_size(self, parallel_mode: ParallelMode, world_size: int):
        self._world_sizes[parallel_mode] = world_size

    def add_group(self, parallel_mode: ParallelMode, group: dist.ProcessGroup) -> int:
        self._groups[parallel_mode] = group

    def get_group(self, parallel_mode: ParallelMode) -> dist.ProcessGroup:
        return self._groups[parallel_mode]

    def add_ranks_in_group(self, parallel_mode: ParallelMode, ranks_in_group: List[int]):
        self._ranks_in_group[parallel_mode] = ranks_in_group

    def get_ranks_in_group(self, parallel_mode: ParallelMode) -> List[int]:
        return dist.get_process_group_ranks(self._groups[parallel_mode])

    def get_next_local_rank(self, rank, parallel_mode: ParallelMode) -> int:
        world_size = self.get_world_size(parallel_mode)
        return (rank + 1) % world_size

    def get_prev_local_rank(self, rank, parallel_mode: ParallelMode) -> int:
        world_size = self.get_world_size(parallel_mode)
        return (rank - 1) % world_size

    def destroy(self):
        assert self.is_initialized(ParallelMode.GLOBAL), "Global group must be initialized before destroying."
        for mode, group in self._groups.items():
            assert self.is_initialized(mode), f"{mode} group must be initialized before destroying."
            if mode is not ParallelMode.GLOBAL:
                dist.barrier()
                dist.destroy_process_group(group)

        dist.barrier()
        dist.destroy_process_group()

        if self.pipeline_parallel_size > 1:
            rpc.shutdown()

        self._groups.clear()



from abc import ABC, abstractclassmethod
from typing import TypedDict

from torch.distributed import ProcessGroup

from pipegoose.distributed.parallel_mode import ParallelMode


class ProcessGroupResult(TypedDict):
    local_rank: int
    local_world_size: int
    process_group: ProcessGroup
    parallel_mode: ParallelMode


class ProcessGroupInitializer(ABC):
    def __init__(
        self, rank: int, world_size: int, tensor_parallel_size: int, pipeline_parallel_size: int, data_parallel_size: int
    ):
        self.rank = rank
        self.world_size = world_size
        self.tensor_parallel_size = tensor_parallel_size
        self.pipeline_parallel_size = pipeline_parallel_size
        self.data_parallel_size = data_parallel_size

    @abstractclassmethod
    def init_dist_group(self) -> ProcessGroupResult:
        raise NotImplementedError



import torch
import torch.distributed.rpc as rpc


def remote_add(x, y):
    return x + y


def run_server():
    rpc.init_rpc(
        "worker0",
        rank=0,
        world_size=2,
        rpc_backend_options=rpc.TensorPipeRpcBackendOptions(
            init_method="tcp://127.0.0.1:29501",
        ),
    )
    print("Server ready")
    rpc.shutdown()


def run_client():
    rpc.init_rpc(
        "worker1",
        rank=1,
        world_size=2,
        rpc_backend_options=rpc.TensorPipeRpcBackendOptions(
            init_method="tcp://127.0.0.1:29501",
        ),
    )
    print("Client ready")
    x = torch.tensor([1, 2, 3])
    y = torch.tensor([4, 5, 6])

    # Use RPC to execute the remote_add function on the server
    result = rpc.rpc_sync("worker0", remote_add, args=(x, y))
    print(f"Client received result: {result}")
    rpc.shutdown()


if __name__ == "__main__":
    import multiprocessing

    # Spawn two processes: one for the server and one for the client
    p1 = multiprocessing.Process(target=run_server)
    p2 = multiprocessing.Process(target=run_client)
    p1.start()
    p2.start()
    p1.join()
    p2.join()



from typing import List

import torch
from torch import nn

from pipegoose.nn.pipeline_parallel.partrition import BasePartitioner


class GPipe:
    def __init__(self, module: nn.Sequential, devices: List[torch.device], partritioner: BasePartitioner) -> None:
        pass



from typing import Any

import torch
from einops import rearrange
from torch import nn
from torchtyping import TensorType

from pipegoose.distributed.functional import all_reduce
from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.distributed.parallel_mode import ParallelMode
from pipegoose.nn.tensor_parallel._utils import get_vocab_range_idx


class _VocabParallelCrossEntropy(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx: Any,
        parallel_logits: TensorType["batch_size", "n_samples", "vocab_size"],
        targets: TensorType["batch_size", "n_samples"],
        parallel_context: ParallelContext,
    ) -> torch.Tensor:
        def normalize_logits(parallel_logits):
            logit_max = torch.max(parallel_logits, dim=-1)[0]
            logit_max = all_reduce(
                logit_max,
                op=torch.distributed.ReduceOp.MAX,
                parallel_context=parallel_context,
                parallel_mode=ParallelMode.TENSOR,
            )
            normalized_parallel_logits = parallel_logits - logit_max.unsqueeze(-1)
            return normalized_parallel_logits

        def get_predicted_logits(parallel_logits, targets):
            rank = parallel_context.get_local_rank(ParallelMode.TENSOR)
            partition_size = parallel_logits.shape[-1]
            vocab_start_idx, vocab_end_idx = get_vocab_range_idx(partition_size, rank)

            target_mask = (targets < vocab_start_idx) | (targets >= vocab_end_idx)
            masked_targets = targets.clone() - vocab_start_idx
            masked_targets[target_mask] = 0

            parallel_logits = rearrange(parallel_logits, "batch_size seq_len vocab_size -> (batch_size seq_len) vocab_size")
            masked_targets_1d = rearrange(masked_targets, "batch_size seq_len -> (batch_size seq_len)")
            predicted_logits = parallel_logits[torch.arange(masked_targets_1d.size(0)), masked_targets_1d]
            predicted_logits = torch.where(
                rearrange(target_mask, "batch_size seq_len -> (batch_size seq_len)") == False, predicted_logits, 0.0
            )
            predicted_logits = all_reduce(
                predicted_logits, parallel_context=parallel_context, parallel_mode=ParallelMode.TENSOR
            )
            return predicted_logits

        # NOTE: parallel cross entropy still works without normalizing logits
        parallel_logits = normalize_logits(parallel_logits)
        predicted_logits = get_predicted_logits(parallel_logits, targets)

        exp_logits = torch.exp(parallel_logits)
        sum_exp_logits = exp_logits.sum(dim=-1)
        sum_exp_logits = all_reduce(sum_exp_logits, parallel_context=parallel_context, parallel_mode=ParallelMode.TENSOR)

        loss = torch.log(sum_exp_logits) - predicted_logits

        return loss

    @staticmethod
    def backward(ctx):
        pass


class VocabParallelCrossEntropy(nn.Module):
    def __init__(self, parallel_context: ParallelContext):
        super().__init__()
        # TODO: support reduce_mean, ignore_index
        self.parallel_context = parallel_context

    def forward(
        self, logits: TensorType["batch_size", "n_samples", "vocab_size"], targets: TensorType["batch_size", "n_samples"]
    ) -> torch.Tensor:
        loss = _VocabParallelCrossEntropy.apply(logits, targets, self.parallel_context)
        loss = loss.mean() / len(targets)
        return loss



import torch.distributed as dist

from pipegoose.distributed._initializers.initializer import (
    ProcessGroupInitializer,
    ProcessGroupResult,
)
from pipegoose.distributed.parallel_mode import ParallelMode


class DataParallelGroupInitializer(ProcessGroupInitializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_pipeline_parallel_groups = self.world_size // self.pipeline_parallel_size

    def init_dist_group(self) -> ProcessGroupResult:
        local_rank = None
        process_group = None
        local_world_size = None
        ranks_in_group = None
        parallel_mode = ParallelMode.DATA

        for i in range(self.pipeline_parallel_size):
            start_rank = i * self.num_pipeline_parallel_groups
            end_rank = (i + 1) * self.num_pipeline_parallel_groups

            for j in range(self.tensor_parallel_size):
                ranks = list(range(start_rank + j, end_rank, self.tensor_parallel_size))
                # NOTE: dist.new_group() must be called collectively by all the processes
                # that would be part of the group, which means every process in the group
                # needs to call this function. If only a subset of the processes call new_group(),
                # it will hang because it's waiting for the rest of the processes to join.
                process_group = dist.new_group(ranks=ranks)

                if self.rank in ranks:
                    local_rank = ranks.index(self.rank)
                    local_world_size = len(ranks)
                    ranks_in_group = ranks

        return {
            "local_rank": local_rank,
            "process_group": process_group,
            "local_world_size": local_world_size,
            "ranks_in_group": ranks_in_group,
            "parallel_mode": parallel_mode,
        }



import torch
import torch.nn.functional as F
from torch import nn

from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.distributed.parallel_mode import ParallelMode
from pipegoose.nn.tensor_parallel._functional import reduce_to_tensor_group
from pipegoose.nn.tensor_parallel._utils import get_vocab_range_idx


class ParallelEmbedding(nn.Module):
    def __init__(self, num_embeddings: int, embedding_dim: int, parallel_context: ParallelContext):
        super().__init__()
        world_size = parallel_context.get_world_size(ParallelMode.TENSOR)

        assert num_embeddings % world_size == 0, "num_embeddings must be divisible by world_size"

        num_embeddings_per_partition = num_embeddings // world_size

        self.parallel_context = parallel_context
        self.weight = nn.Parameter(torch.randn(num_embeddings_per_partition, embedding_dim))
        self.vocab_start_idx, self.vocab_end_idx = get_vocab_range_idx(
            num_embeddings_per_partition, rank=parallel_context.get_local_rank(ParallelMode.TENSOR)
        )
        self.world_size = world_size

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.world_size > 1:
            input_mask = (input < self.vocab_start_idx) | (input >= self.vocab_end_idx)
            # align global embedding indices to local embedding indices
            masked_input = input.clone() - self.vocab_start_idx
            masked_input[input_mask] = 0
        else:
            masked_input = input

        parallel_output = F.embedding(masked_input, self.weight)

        if self.world_size > 1:
            parallel_output[input_mask, :] = 0.0

        output = reduce_to_tensor_group(parallel_output, parallel_context=self.parallel_context)

        return output



import copy
from typing import Dict, List

import torch

from pipegoose.distributed.parallel_context import ParallelContext


class ParameterSharding:
    def __init__(self, param_groups: List[Dict[str, torch.Tensor]], parallel_context: ParallelContext):
        self.param_groups = param_groups
        self.parallel_context = parallel_context

    def shard(self) -> List[Dict[str, torch.Tensor]]:
        world_size = self.parallel_context.get_world_size()

        partition_parameters = [[] for _ in range(world_size)]
        sizes = [0 for _ in range(world_size)]

        for param_group in self.param_groups:
            param_lists = [[] for _ in range(world_size)]

            for param in param_group["params"]:
                next_rank = sizes.index(min(sizes))
                param_lists[next_rank].append(param)
                sizes[next_rank] += param.numel()

            for rank, params in enumerate(param_lists):
                partitioned_param_group = copy.copy(param_group)
                partitioned_param_group["params"] = params
                partition_parameters[rank].append(partitioned_param_group)

        return partition_parameters



import torch
from torch import nn


class ParamHook:
    def __init__(self, module: nn.Module):
        self.module = module

    def register_pre_backward_hook(self):
        if not torch.is_grad_enabled():
            return

        for p in self.module.parameters():
            if p.requires_grad:
                p_temp = p.expand_as(p)
                assert p_temp.grad_fn is not None

    def register_post_backward_hook(self):
        pass



from typing import Tuple

from torch import nn


def is_linear_parallelizable(module: nn.Module) -> bool:
    pass


def _is_column_parallel(module: nn.Module) -> bool:
    pass


def _is_row_parallel(module: nn.Module) -> bool:
    pass


def get_vocab_range_idx(partition_size: int, rank: int) -> Tuple[int, int]:
    start_idx = rank * partition_size
    end_idx = start_idx + partition_size
    return start_idx, end_idx



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from threading import Thread

from pipegoose.nn.pipeline_parallel2.batch import Batch


class JobStatus(Enum):
    EXECUTING = auto()
    SUCCEED = auto()


@dataclass
class Metadata:
    # data
    microbatch_idx: int
    partition_idx: int

    # job
    is_forward: bool
    is_training: bool
    is_grad_enabled: bool  # not implemented yet
    is_fp16: bool  # not implemented yet
    status: JobStatus

    # communication
    src: int
    dst: int


class Job(ABC):
    def __init__(self, input: Batch, meta: Metadata) -> None:
        self._meta = meta
        self.input = input

    @abstractmethod
    def compute(self):
        raise NotImplementedError("not implemented")

    def meta(self):
        return self._meta


class ForwardJob(Job):
    pass


class BackwardJob(Job):
    pass


class JobRegister(ABC):
    def __init__(self, data):
        pass


class JobSelector(Thread):
    pass



from queue import Queue
from threading import Thread
from typing import Callable, List

from pipegoose.constants import PIPELINE_MAX_WORKERS, PIPELINE_MIN_WORKERS
from pipegoose.nn.pipeline_parallel2._utils import sleep


class Worker(Thread):
    def __init__(self, selected_jobs: Queue, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._selected_jobs = selected_jobs
        # self._running = False

    @property
    def running(self) -> bool:
        # return self._running
        return self.is_alive()

    def run(self):
        while True:
            # TODO: separate alive and running
            job = self._selected_jobs.get()
            # self._running = True

            job.compute()

            # self._running = False


class WorkerPoolWatcher(Thread):
    def __init__(self, worker_pool: List[Worker], min_workers: int, max_workers: int, spawn_worker: Callable, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.worker_pool = worker_pool
        self.min_workers = min_workers
        self.max_workers = max_workers
        self.spawn_worker = spawn_worker

    def run(self):
        while True:
            num_working = self._num_working_workers()

            # TODO: should we spawn more than initial workers?
            # TODO: fix race condition in some idle threads
            if num_working < self.max_workers:
                self.spawn_worker()

    def _num_working_workers(self) -> int:
        num_working = 0
        for worker in self.worker_pool:
            num_working += 1 if worker.running is True else 0
        return num_working


class JobSelector(Thread):
    def __init__(self, pending_jobs: Queue, selected_jobs: Queue, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pending_job = pending_jobs
        self._selected_jobs = selected_jobs

    def run(self):
        while True:
            job = self._select_job()
            self._selected_jobs.put(job)

    def _select_job(self):
        while len(self._pending_job) < 1:
            sleep()

            job = list(sorted(self._pending_job))[0]
            self._pending_job.remove(job)
            return job


class WorkerManager(Thread):
    def __init__(
        self,
        num_workers: int = PIPELINE_MIN_WORKERS,
        min_workers: int = PIPELINE_MIN_WORKERS,
        max_workers: int = PIPELINE_MAX_WORKERS,
        pending_jobs: Queue = Queue(),
        selected_jobs: Queue = Queue(),
    ):
        # job just created but not yet selected
        self._pending_jobs = pending_jobs
        # job selected to be executed
        self._selected_jobs = selected_jobs
        self._worker_pool = []
        self.num_workers = num_workers
        self.min_workers = min_workers
        self.max_workers = max_workers

    @property
    def pending_jobs(self) -> Queue:
        return self._pending_jobs

    @property
    def selected_jobs(self) -> Queue:
        return self._selected_jobs

    @property
    def worker_pool(self) -> List[Worker]:
        return self._worker_pool

    def _spawn_job_selector(self):
        job_selector = JobSelector(self._pending_jobs, self._selected_jobs)
        job_selector.setDaemon(True)
        job_selector.start()

        self.job_selector = job_selector

    def _spawn_pool_watcher(self):
        pool_watcher = WorkerPoolWatcher(
            self.worker_pool, min_workers=self.min_workers, max_workers=self.max_workers, spawn_worker=self._spawn_a_worker
        )
        pool_watcher.setDaemon(True)
        pool_watcher.start()

        # TODO: delete after testing
        self.pool_watcher = pool_watcher

    def _spawn_a_worker(self):
        worker = Worker(selected_jobs=self._selected_jobs)
        worker.setDaemon(True)
        worker.start()
        self._worker_pool.append(worker)

    def _spawn_initial_workers(self):
        for _ in range(self.num_workers):
            self._spawn_a_worker()

    def spawn(self):
        self._spawn_job_selector()
        self._spawn_initial_workers()
        self._spawn_pool_watcher()

    def destroy(self):
        # TODO: why can't we join() here?
        for worker in self.worker_pool:
            # TODO: wait for workers finish their jobs
            # before joining
            worker.join()
        # self.worker_pool.remove(worker)



# from abc import ABC, abstractclassmethod


import torch
from torch import nn

from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.nn.tensor_parallel.layers import (
    ParallelizeAttention,
    ParallelizeEmbedding,
    ParallelizeLayerNorm,
    ParallelizeLinear,
)

# from pipegoose.distributed.parallel_mode import ParallelMode


# class ParallelizeModule(ABC):
#     def __init__(self, module: nn.Module, parallel_context: ParallelContext):
#         self.module = module
#         self.parallel_context = parallel_context

#     @abstractclassmethod
#     def parallelize(self):
#         raise NotImplementedError

#     @abstractclassmethod
#     def deparallelize(self):
#         raise NotImplementedError


# class ParallelizeLinear(ParallelizeModule):
#     def parallelize(self):
#         # tensor_parallel_size = self.parallel_context.get_config(
#         #     mode=ParallelMode.TENSOR
#         # )
#         pass

#     def deparallelize(self):
#         pass


# class ParallelizeEmbedding(ParallelizeModule):
#     def parallelize(self):
#         vocab_size, embedding_size = self.module.weight.size()


# class ParallelizeLayerNorm(ParallelizeModule):
#     pass


# class ParallelizeAttention(ParallelizeModule):
#     pass


class TensorParallel:
    """Turn a sequential model into a tensor-parallel model.

    Inspired by OSLO's TensorParallel: https://github.com/EleutherAI/oslo/blob/00e3be56446df37a0372a93a094863ffc89a2f8b/oslo/torch/nn/parallel/tensor_parallel/tensor_parallel.py#L51
    """

    def __init__(self, module: nn.Module, parallel_context: ParallelContext):
        super().__init__()
        self.module = module
        self.parallel_context = parallel_context

    @torch.no_grad()
    def parallelize(self):
        paralleler = {
            "linear": ParallelizeLinear,
            "embedding": ParallelizeEmbedding,
            "layer_norm": ParallelizeLayerNorm,
            "attention": ParallelizeAttention,
        }

        for name, module in self.module.named_modules():
            if name in paralleler:
                paralleler[name](module, self.parallel_context).parallelize()

    def _parallelize_embedding(self):
        for module_name, module in self.module.named_modules():
            if isinstance(module, nn.Embedding):
                pass

    def _parallelize_linear(self):
        pass

    def _parallize_layernorm(self):
        for _, module in self.module.named_modules():
            if isinstance(module, nn.LayerNorm):
                pass

    def _resize_vocab_size(self, module: nn.Module):
        pass



from abc import ABC, abstractclassmethod

from torch import nn

from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.distributed.parallel_mode import ParallelMode


class ParallelizeModule(ABC):
    def __init__(self, module: nn.Module, parallel_context: ParallelContext):
        self.module = module
        self.parallel_context = parallel_context

    @abstractclassmethod
    def parallelize(self):
        raise NotImplementedError

    @abstractclassmethod
    def deparallelize(self):
        raise NotImplementedError


class ParallelizeLinear(ParallelizeModule):
    def parallelize(self):
        module = self.module
        parallel_context = self.parallel_context

        local_rank = parallel_context.get_local_rank(ParallelMode.TENSOR)
        local_world_size = parallel_context.get_local_world_size(ParallelMode.TENSOR)

        input_size, output_size = module.weight.size()
        output_per_partition = output_size // local_world_size

        start_element = local_rank * output_per_partition
        end_element = (local_rank + 1) * output_per_partition

        weight = module.weight.data
        module.weight.data = weight[:, start_element:end_element]

        if module.bias is not None:
            bias = module.bias.data
            module.bias.data = bias[start_element:end_element]

        return module

    def deparallelize(self):
        pass


class ParallelizeEmbedding(ParallelizeModule):
    def parallelize(self):
        vocab_size, embedding_size = self.module.weight.size()


class ParallelizeLayerNorm(ParallelizeModule):
    pass



from typing import Any, Tuple

import torch


class StartDependency(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        phony = torch.empty(1, requires_grad=False, device=input.device)
        return input.detach(), phony.detach()

    @staticmethod
    def backward(ctx, grad_input: torch.Tensor, grad_phony: torch.Tensor) -> torch.Tensor:
        return grad_input


class EndDependency(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor, phony: torch.Tensor) -> torch.Tensor:
        return input.detach()

    @staticmethod
    def backward(ctx: Any, grad_input: torch.Tensor) -> torch.Tensor:
        return grad_input, None


def create_backward_dependency(source_tensor: torch.Tensor, target_tensor: torch.Tensor):
    source_tensor, phony = StartDependency.apply(source_tensor)
    target_tensor = EndDependency.apply(target_tensor, phony)

    return source_tensor, target_tensor


class Wait(torch.autograd.Function):
    @staticmethod
    def forward(ctx, prev_stream, next_stream, input: torch.Tensor) -> torch.Tensor:
        ctx.prev_stream = prev_stream
        ctx.next_stream = next_stream



from abc import ABC, abstractclassmethod

import torch
import torch.nn.functional as F
from torch import nn

from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.nn.data_parallel.utils import free_storage


class ShardingStategy(ABC):
    @abstractclassmethod
    def shard(self):
        raise NotImplementedError("You must implement shard method")


class GreedySharding(ShardingStategy):
    def __init__(self, module: nn.Module, parallel_context: ParallelContext):
        self.module = module
        self.parallel_context = parallel_context

    @torch.no_grad()
    def shard(self) -> nn.Module:
        module = self.module

        self._shard_parameters()

        for name, param in module.named_parameters():
            assert hasattr(param, "_is_sharded"), f"{name} is haven't sharded"

        return module

    def _shard_parameters(self) -> torch.Tensor:
        module = self.module
        world_size = self.parallel_context.get_world_size()

        for p in module.parameters():
            assert not hasattr(p, "_is_sharded")

            if world_size > 1:
                orig_data = p.data
                p.data = self._get_shard(p.data)
                p._is_sharded = True
                free_storage(orig_data)

    def _get_shard(self, data: torch.Tensor) -> torch.Tensor:
        world_size = self.parallel_context.get_world_size()
        rank = self.parallel_context.get_rank()

        chunks = list(data.flatten().chunk(world_size))

        while len(chunks) < world_size:
            chunks.append(torch.empty(0))

        shard = chunks[rank].clone()
        num_to_pad = chunks[0].numel() - shard.numel()
        if num_to_pad > 0:
            shard = F.pad(shard, [0, num_to_pad])
        return shard



from enum import Enum, auto
from typing import Any, Callable

from pipegoose.nn.pipeline_parallel.microbatch import Batch


class BaseTask:
    def __call__(self, *args, **kwargs) -> Any:
        return self.compute(*args, **kwargs)


class TaskStatus(Enum):
    DONE = auto()
    NOT_DONE = auto()


class Task:
    def __init__(self, compute: Callable[[], Batch], is_done: TaskStatus = TaskStatus.NOT_DONE):
        self._compute = compute
        self.is_done = is_done

    def compute(self) -> Batch:
        return self._compute()



from typing import Tuple


class TensorParallelInformation:
    def __init__(self, module_name: Tuple[str], **kwargs):
        self.module_name = module_name
        self.kwargs = kwargs


class Column(TensorParallelInformation):
    pass


class Row(TensorParallelInformation):
    pass


class ParallelMapping:
    # TODO: make this extendable
    # so user can define their own mapping

    __MAPPING__ = {
        "albert-base-v2": [Column(("query", "key", "value")), Row("attention.dense")],
        "bloom-560m": [
            Column(("dense_h_to_4h",)),
            Row(("dense_4h_to_h",)),
        ],
    }

    def _search(cls, module_name: str) -> TensorParallelInformation:
        """
        Search for module_name in mappings.
        """
        for _, items in cls.__MAPPING__.items():
            for item in items:
                if module_name in item.module_name:
                    return item
        return None

    @classmethod
    def is_column_parallel(cls, module_name: str) -> bool:
        item = cls._search(module_name)
        if item is None:
            return False
        return isinstance(item, Column)

    @classmethod
    def is_row_parallel(cls, module_name: str) -> bool:
        item = cls._search(module_name)
        if item is None:
            return False
        return isinstance(item, Row)



from typing import Any, Dict, Iterable, List

import torch
from torch import nn
from torch.optim import Optimizer

from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.optim import DistributedOptimizer
from pipegoose.optim.zero.sharding import ParameterSharding


class ZeroRedundancyOptimizer(DistributedOptimizer):
    def __init__(
        self,
        params: Iterable[List[nn.Parameter]],
        optim: Optimizer,
        parallel_context: ParallelContext,
        sharding: ParameterSharding,
        **default: Any
    ):
        # TODO: accept params, and optim separately
        self.params = params
        self._optim_constructor = optim
        self.parallel_context = parallel_context
        self.default = default

        # world_size = parallel_context.get_world_size(ParallelMode.GLOBAL)

    def _partition_and_move_to_rank(self):
        self._partition_params(self.optim.param_groups)

    def _construct_local_optim(self, local_params: Dict[str, torch.Tensor]):
        self.optim = self._optim_constructor(local_params, **self.default)

    def _sync_hyperparams(self, source: List[Dict[Any, Any]], destination: List[Dict[Any, Any]]):
        for source_group, destination_group in zip(source, destination):
            for k in source_group.keys():
                if k != "params":
                    destination_group[k] = source_group[k]

    def setup(self):
        pass

    def step(self, **kwargs):
        self._sync_hyperparams(self.param_groups, self.optim.param_groups)

        self.optim.step(**kwargs)



from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from pipegoose.distributed.parallel_context import ParallelContext
from pipegoose.distributed.parallel_mode import ParallelMode
from pipegoose.nn.tensor_parallel._functional import (
    broadcast_to_tensor_group,
    gather_to_tensor_group,
    reduce_to_tensor_group,
    scatter_to_tensor_group,
)


class ColumnParallelLinear(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        gather_output: bool = False,
        parallel_context: Optional[ParallelContext] = None,
    ):
        super().__init__()
        out_per_partition = self._get_output_per_partition(out_features, parallel_context)

        self.gather_output = gather_output
        self.parallel_context = parallel_context
        self.weight = nn.Parameter(torch.randn(out_per_partition, in_features))

        if bias is True:
            self.bias = nn.Parameter(torch.randn(out_per_partition))

    def _get_output_per_partition(self, out_features: int, parallel_context: ParallelContext) -> int:
        local_world_size = parallel_context.get_world_size(ParallelMode.TENSOR)
        return out_features // local_world_size

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        input_parallel = broadcast_to_tensor_group(input, self.parallel_context)
        outputs = F.linear(input_parallel, self.weight)

        if self.bias is not None:
            outputs = outputs + self.bias

        if self.gather_output:
            outputs = gather_to_tensor_group(outputs, dim=-1, parallel_context=self.parallel_context)

        return outputs


class RowParallelLinear(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        parallel_context: Optional[ParallelContext] = None,
    ) -> None:
        super().__init__()
        in_per_partition = self._get_input_per_partition(in_features, parallel_context)

        self.in_features = in_features
        self.out_features = out_features
        self.parallel_context = parallel_context

        self.weight = nn.Parameter(torch.randn(out_features, in_per_partition))

        if bias is True:
            self.bias = nn.Parameter(torch.randn(out_features))

    def _get_input_per_partition(self, in_features: int, parallel_context: ParallelContext) -> int:
        local_world_size = parallel_context.get_world_size(ParallelMode.TENSOR)
        return in_features // local_world_size

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        input_parallel = scatter_to_tensor_group(input, dim=-1, parallel_context=self.parallel_context)
        output_parallel = F.linear(input_parallel, self.weight)
        outputs = reduce_to_tensor_group(output_parallel, parallel_context=self.parallel_context)

        if self.bias is not None:
            outputs = outputs + self.bias

        return outputs



from typing import Tuple

import torch

from pipegoose.nn.pipeline_parallel.stream import (
    StreamType,
    default_stream,
    get_device,
    record_stream,
    use_stream,
)


class Context:
    prev_stream: torch.cuda.Stream
    next_stream: torch.cuda.Stream


class Copy(torch.autograd.Function):
    """Synchronous dat transfer between streams."""

    @staticmethod
    def forward(ctx: Context, prev_stream: StreamType, next_stream: StreamType, input: torch.Tensor) -> torch.Tensor:
        ctx.prev_stream = prev_stream
        ctx.next_stream = next_stream

        compute_stream = default_stream(get_device(next_stream))

        with use_stream(prev_stream), use_stream(next_stream):
            moved_input = input.to(get_device(next_stream))
            record_stream(input, prev_stream)
            record_stream(moved_input, compute_stream)

        return moved_input

    @staticmethod
    def backward(ctx: Context, grad_input: torch.Tensor) -> Tuple[None, None, torch.Tensor]:
        prev_stream = ctx.prev_stream
        next_stream = ctx.next_stream

        compute_stream = default_stream(get_device(prev_stream))

        with use_stream(prev_stream), use_stream(next_stream):
            moved_grad_input = grad_input.to(get_device(prev_stream))

            record_stream(grad_input, next_stream)
            record_stream(moved_grad_input, compute_stream)

        return tuple([None, None, grad_input])
